#!/usr/bin/python3
# ::::: __________________________________________________________________ :::::
# : ____\ ._ ____ _____ __. ____ ___ _______ .__ ______ .__ _____ .__ _. /____ :
# __\ .___! _\__/__    / _|__   / _/_____  __|  \ gRK __|_ \  __  |_ \ !___. /__
# \   ! ___/  |/  /___/  |   \__\ ._/  __\/  \   \___/  |/  \/  \_./  \___ !   /
# /__  (___   /\____\____|\   ____|   /  /___|\   ______.    ____\|\   ___)  __\
#   /____  \_/ ___________ \_/ __ |__/ _______ \_/ ____ |___/ _____ \_/  ____\
# :     /________________________________________________________________\     :
# :::::       +  p  H  E  N  O  M  p  R  O  D  U  C  T  I  O  N  S  +      :::::
# ==============================================================================
#
# -----------------------------------------
# - modName: mrc_client multiplexer       -
# - majorVersion: 1.3                     -
# - minorVersion: 7                       -
# - author: Stackfault                    -
# - publisher: Phenom Productions         -
# - website: https://www.phenomprod.com   -
# - email: stackfault@bottomlessabyss.net -
# - bbs: bbs.bottomlessabyss.net:2023     -
# -----------------------------------------
#
# Based on previous work from Gryphon of Cyberia BBS
#
# The code have been completely reviewed and improved everywhere I could see
# room for improvement without breaking compatibility.
#
# Major changes:
#
# v1.2.5
#
# - Error trapping on all critical locations
# - Socket routine rewrite and now non-blocking
# - Internal auto-restart, no need for an external restart script
# - New commands added and supported by the new server code
# - Message serialization allowing very fast message rate and proper display order
# - Bonus server stats data allowing an in-bbs applet to show MRC status (See samples)
# - Graceful client shutdown notification to the server
# - New BBS information subsystem allowing BBS to provide connection info details
# - New startup check to allow for smoother installation and configuration
#
# v1.2.7
#
# - Update available/Client too old notifications and validation
# - Client latency reporting
# - Added support for upcoming server activity level stats
# - Increase both frequency and tolerance of stats reporting
# - Other smaller fixes
#
# v1.2.9
# - Improved handling of incomplete and invalid packets
# - Implemented client stats transfer
# - Increased tcp buffer size
# - Improved the stats reporting
# - Rebased the message serialization
#
# v1.3.3
# - Python 3 support (2 different scripts)
# - Server is no longer on command-line and now in mrc_config.py
# - SSL support
# - CAPABILITIES support (in preparation for future use)
# - Duplicate process check via pidfile
#
# v1.3.4
# - Minor bug fixes
#
# Make sure to use the new mrc_config.py so you can take advantage of some new
# features.
#
# v1.3.5
#
# - Fixed minor issues in template for Windows hosts
# - Fixed minor issues while run on a Windows host
#
# v1.3.6
#
# - Single script now supported both by Python 2.7.18 and 3.6+
#
# v1.3.7
#
# - Version bump


import os, sys, fnmatch, glob, re, hashlib
import time, socket, errno, platform, traceback

# Import site config
from mrc_config import *

# Pre-Flight checks
try:
    os.chdir(bbspath)
except:
    print("Please configure BBS Path in mrc_config.py")
    sys.exit(1)

if "Unconfigured BBS" in [
    bbsname,
    info_web,
    info_telnet,
    info_ssh,
    info_sysop,
    info_desc,
]:
    print("Please configure your BBS details in mrc_config.py")
    sys.exit(1)

msleep = lambda x: time.sleep(x / 1000.0)

# Change this info
tempdir = "%s%stemp" % (bbspath, os.sep)
datadir = "%s%sdata" % (bbspath, os.sep)
chatdats = "%s%schat*.dat" % (datadir, os.sep)

# Align this path with the MRC MPL (Default: {mrcdatadir}/mrc)
mrcdir = "%s%smrc" % (datadir, os.sep)

# Platform information
version = "1.3.7"
platform_name = "MYSTIC"
system_name = platform.system()
machine_arch = platform.machine()
debugflag = False
version_string = "%s/%s.%s/%s" % (platform_name, system_name, machine_arch, version)
client_version = "Multi Relay Chat Client v%s 2024-09-01 [sf]" % version
capabilities = "MCI CTCP SSL"

# Global vars
intv = [1, 5, 10, 30, 60, 90, 120, 180, 240, 300]  # Auto-restart intervals
timebase = int(time.time())
registry = {}
mrcstats = ""
mrchash = ""


# Console logger
def logger(loginfo):
    ltime = time.asctime(time.localtime(time.time()))
    print("%s  %s" % (ltime, loginfo.strip()))
    sys.stdout.flush()


# Check if already running
def checkpid(c):
    # Check Pid
    if c == 0:
        try:
            p = open(pidfile, "r")
            pid = p.read()
            p.close()
            os.kill(int(pid), 0)
        except:
            return False
        else:
            return True

    # Create pid file
    if c == 1:
        try:
            p = open(pidfile, "w")
            p.write(str(os.getpid()))
            p.close()
            return
        except:
            logger("Error creating pid file")
            sys.exit()

    # Delete pid file
    if c == 2:
        try:
            os.remove(pidfile)
            return
        except:
            logger("Error deleting pid file")


# Strip MCI color codes
def stripmci(text):
    return re.sub(r"\|[0-9]{2}", "", text)


# User chatlog for DLCHATLOG
def chatlog(data):
    if "CLIENT~" not in data and "SERVER~" not in data:
        ltime = time.asctime(time.localtime(time.time()))
        packet = data.split("~")
        message = stripmci(packet[6])
        clogfile = "%s%smrcchat.log" % (mrcdir, os.sep)
        clog = open(clogfile, "a")
        clog.write("%s %s\n" % (ltime, message))
        clog.close()


# Socket sender to server
def send_server(data):
    global registry
    global mrcserver
    if data:
        try:
            regstp = int(time.time() * 1000)
            regstr = data.strip()
            registry[regstr] = regstp
            mrcserver.send(data.encode())
        except:
            logger("Connection error")
            try:
                mrcserver.shutdown(2)
            except:
                pass
            finally:
                mrcserver.close()


# Temp files cleaning routine
def clean_files():
    mrcfiles = os.listdir(mrcdir)
    for file in mrcfiles:
        if fnmatch.fnmatch(file, "*.mrc"):
            mrcfile = "%s%s%s" % (mrcdir, os.sep, file)
            os.remove(mrcfile)


# Read queued file from MRC, ignoring stale files older than 10s
def send_mrc():
    mrcfiles = os.listdir(mrcdir)
    for file in mrcfiles:
        if fnmatch.fnmatch(file, "*.mrc"):
            mrcfile = "%s%s%s" % (mrcdir, os.sep, file)
            ft = os.path.getmtime(mrcfile)

            # Do not forward packets older than 10s
            if time.time() - ft < 10:
                try:
                    # Avoid reading a file still open by the MPL by
                    # opening it read-write to check for locking
                    f = open(mrcfile, "r+")
                    fl = f.readline()
                    f.close()

                    if fl.count("~") > 5:
                        mline = fl.split("~")
                        fromuser = mline[0]
                        frombbs = mline[1]
                        touser = mline[3]
                        message = mline[6]
                        if message == "VERSION":
                            deliver_mrc(
                                "CLIENT~~~%s~%s~~|07- %s~"
                                % (fromuser, frombbs, client_version)
                            )
                            send_server(fl)

                        elif touser == "CLIENT" and message == "LATENCY":
                            deliver_mrc(
                                "SERVER~~~CLIENT~%s~~LATENCY:%s~" % (frombbs, latency)
                            )

                        elif touser == "CLIENT" and message == "STATS":
                            deliver_mrc(
                                "SERVER~~~CLIENT~%s~~STATS:%s~" % (frombbs, mrcstats)
                            )

                        else:
                            send_server(fl)

                        os.remove(mrcfile)
                    else:
                        if debugflag:
                            logger("File %s contains invalid packet" % mrcfile)

                except IOError:
                    if debugflag:
                        logger("MRC file still busy")
                    pass
                except:
                    if debugflag:
                        logger("Error:" + traceback.print_exc())
            else:
                os.remove(mrcfile)


# Write time serialized file for display in MRC
def deliver_mrc(server_data):
    global registry
    global latency
    global mrcstats
    global timebase

    curtime = int(time.time() * 1000)
    if server_data.strip() in list(registry.keys()):
        pkttime = int(registry[server_data.strip()])
        roundtrip = curtime - pkttime
        if roundtrip > 1:
            latency = roundtrip
            if debugflag:
                logger(
                    "LATENCY: Current: %s - Packet: %s = RoundTrip: %s"
                    % (curtime, pkttime, roundtrip)
                )
        registry.clear()

    # Make up a serialized filename based on time
    fileserial = int((time.time() - timebase) * 1000)

    # Wrap the file serial if longer than 8 chars
    if fileserial > 99999999:
        timebase = int(time.time())
        fileserial = int((time.time() - timebase) * 1000)

    # Zeropad the filename
    filename = "%08d.mrc" % fileserial

    try:
        packet = server_data.split("~")
        fromuser = packet[0]
        fromsite = packet[1]
        fromroom = packet[2]
        touser = packet[3]
        tosite = packet[4]
        toroom = packet[5]
        message = packet[6]
    except:
        logger("Bad packet: %s" % server_data)
        return

    # Manage server PINGs
    if fromuser == "SERVER" and message.lower() == "ping":
        send_im_alive()

    # Manage update available notifications
    elif fromuser == "SERVER" and message.startswith("NEWUPDATE:"):
        logger("Upgrade is available, consider upgrading at your earliest convenience")
        logger("You are using version %s" % version)
        logger("Latest version is %s" % message.split(":")[1])

    # Manage old clients
    elif fromuser == "SERVER" and message.startswith("OLDVERSION:"):
        logger("Your client is too old and can no longer be used.")
        logger("You are using version %s" % version)
        logger("Latest version is %s" % message.split(":")[1])
        raise KeyboardInterrupt

    else:

        # Manage server stats
        if fromuser == "SERVER" and message.startswith("STATS:"):
            statsfile = "%s%smrcstats.dat" % (mrcdir, os.sep)
            try:
                f = open(statsfile, "w")
                f.write(message.split(":")[1])
                f.close()
                mrcstats = message.split(":")[1]
            except:
                logger("Cannot write server stats to %s" % statsfile)

        chatlog(server_data)
        for f in glob.iglob(chatdats):
            if not "chatroom" in f:
                chatfile = "%s%schat" % (datadir, os.sep)
                xy = f.replace(chatfile, tempdir)
                xy = xy[:-4]
                inusefile = "%s%stchat.inuse" % (xy, os.sep)
                if os.path.isfile(inusefile):
                    mrcfile = "%s%s%s" % (xy, os.sep, filename)
                    openfile = open(mrcfile, "a")
                    openfile.write(server_data)
                    openfile.close()
                    msleep(5)


# Respond to server PING
def send_im_alive():
    data = "CLIENT~%s~%s~SERVER~~~IMALIVE:%s~\n" % (bbsname, str(os.getpid()), bbsname)
    send_server(data)


# Send capabilities
def send_capabilities():
    data = "CLIENT~%s~%s~SERVER~~~CAPABILITIES:%s~\n" % (bbsname, mrchash, capabilities)
    send_server(data)


# Send graceful shutdown request to server when exited
def send_shutdown():
    data = "CLIENT~%s~~SERVER~~~SHUTDOWN~\n" % bbsname
    send_server(data)


# Request server stats for applet
def request_stats():
    data = "CLIENT~%s~~SERVER~~~STATS~\n" % bbsname
    send_server(data)


# Send BBS additional info for INFO command
def send_bbsinfo():
    prefix = "CLIENT~%s~~SERVER~ALL~~" % bbsname
    packet = prefix + "INFOWEB:%s~\n" % info_web
    packet += prefix + "INFOTEL:%s~\n" % info_telnet
    packet += prefix + "INFOSSH:%s~\n" % info_ssh
    packet += prefix + "INFOSYS:%s~\n" % info_sysop
    packet += prefix + "INFODSC:%s~\n" % info_desc
    send_server(packet)


# Handle different line separator scenarios
def check_separator(data):
    if data.count("\r\n"):
        return "\r\n"
    elif data.count("\n\r"):
        return "\n\r"
    elif data.count("\r"):
        return "\r"
    else:
        return "\n"


# Main process loop
def mainproc():
    global delay
    global mrcserver
    global latency
    global sock
    global context

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(5)

    restart = 0
    readbuffer = ""
    tdat = ""
    latency = 0

    # Non-blocking socket loop to improve speed
    try:
        trans = "Plain"
        if usessl:
            mrcserver = context.wrap_socket(sock, server_hostname=host)
            trans = "SSL"
            port = sslport
        else:
            # Use plain socket
            mrcserver = sock
            port = plainport

        logger(
            "Attempting connection to %s on port %d using %s transport\n"
            % (host, port, trans)
        )

        mrcserver.connect((host, port))
        mrcserver.setblocking(0)
        handshake = "%s~%s" % (bbsname, version_string)
        mrcserver.send(handshake.encode())

        logger(
            "Connected (%s) to Multi Relay Chat host %s port %d" % (trans, host, port)
        )

        delay = 0
    except:
        logger("Unable to connect to %s:%d" % (host, port))
        return

    send_bbsinfo()
    send_im_alive()
    send_capabilities()

    loop = 1800
    while True:
        msleep(10)
        send_mrc()

        loop += 1

        # Request stats every 20 seconds
        if loop > 2000:
            request_stats()
            loop = 0

        # Here we need two methods for socket, with and without SSL
        # because of failure with the SSL import potentially
        # If we use SSL, we can rely on ssl.SSLWantReadError
        if usessl:

            try:
                readbuffer = mrcserver.recv(16384).decode()

            # Handle SSL socket error
            except ssl.SSLWantReadError:
                continue

            # Handle socket error
            except socket.error as e:
                err = e.args[0]
                if err == errno.EAGAIN or err == errno.EWOULDBLOCK:
                    continue
                else:
                    restart = 1

        # Without SSL, we only rely on socket
        else:
            try:
                readbuffer = mrcserver.recv(16384).decode()

            # Handle socket error
            except socket.error as e:
                err = e.args[0]
                if err == errno.EAGAIN or err == errno.EWOULDBLOCK:
                    continue
                else:
                    restart = 1

        # If we have buffer, we process
        if readbuffer:
            sep = check_separator(readbuffer)
            tdat = readbuffer.split(sep)
            for data in tdat:
                if data:
                    deliver_mrc(data)

        # Handle socket restarts with socket shutdowns
        if restart:
            logger("Lost connection to server\n")
            try:
                mrcserver.shutdown(2)
            finally:
                mrcserver.close()
            return


# Some validation of config to ensure smoother operation
def check_startup():
    global mrchash
    failed = 0

    if not os.path.exists("%s%susers.dat" % (datadir, os.sep)):
        print("The MRC Multiplexer must be started from your BBS directory")
        sys.exit()

    if not os.path.exists(mrcdir):
        os.makedirs(mrcdir)

    if checkpid(0):
        logger("MRC Client is already running, pid file %s" % pidfile)
        sys.exit()

    if len(stripmci(bbsname)) < 5:
        print("Config: 'bbsname' should be set to something sensible")
        failed = 1

    if len(stripmci(bbsname)) > 40:
        print(
            "Config: 'bbsname' cannot be longer than 40 characters after PIPE codes evaluation"
        )
        failed = 1

    for param in ["info_web" "info_telnet", "info_ssh", "info_sysop", "info_desc"]:
        if len(stripmci(param)) > 64:
            print(
                "Config: '%s' cannot be longer than 64 characters after PIPE codes evaluation"
                % param
            )
            failed = 1

    for param in ["info_web" "info_telnet", "info_ssh", "info_sysop", "info_desc"]:
        if len(param) > 128:
            print(
                "Config: '%s' cannot be longer than 128 characters including PIPE codes"
                % param
            )
            failed = 1

    if failed:
        print("This must be fixed in mrc_config.py")
        sys.exit()

    h = open(__file__, "r")
    mrchash = hashlib.sha256(h.read().encode()).hexdigest()
    h.close()


# Main loop
if __name__ == "__main__":
    logger(client_version)

    # Check if we can use SSL on this system
    if usessl:
        try:
            import ssl

            context = ssl.create_default_context()
            if nocertcheck:
                context.check_hostname = False
                context.verify_mode = ssl.CERT_NONE
            port = sslport
        except:
            logger("! Unable to use SSL, reverting to plain\n")
            usessl = 0
            port = plainport

    check_startup()
    checkpid(1)
    delay = 0
    while True:
        try:
            mainproc()

            # Incremental auto-restart built-in
            logger("Reconnecting in %d seconds" % intv[delay])
            time.sleep(intv[delay])
            delay += 1
            if delay > 9:
                delay = 0

        except KeyboardInterrupt:
            logger("Shutting down")
            try:
                send_shutdown()
                try:
                    mrcserver.shutdown(2)
                except:
                    pass
            finally:
                mrcserver.close()

            checkpid(2)
            sys.exit()
        except:
            continue



"""Regression test for a real gap found in a full echomail-subsystem
audit: nothing guarded against two overlapping poll attempts for the
SAME network -- a sysop double-clicking "Poll Now"
(echomail_admin.py's poll_now(), which spawns a bare daemon thread with
no dedup check at all) or a manual poll landing while the scheduled
_poller_loop's own tick for that network is already mid-flight could
run two concurrent BinkP sessions to the same peer, risking duplicate
sends and interleaved hold-queue/ack-gated-stamping writes.

Fixed by reusing the existing status='running' EchomailPollLog row
(already written for poll-in-progress visibility) as a dedup signal:
_do_poll() now checks for an existing running row for the same
network before starting a second one.
"""
import os
import sys
import unittest
from datetime import datetime, timedelta
from pathlib import Path
from unittest.mock import patch

sys.path.insert(0, str(Path(__file__).resolve().parent.parent))

import anetbbs.config as cfg_mod


class PollerConcurrentPollGuardTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._orig_db_uri = cfg_mod.TestingConfig.SQLALCHEMY_DATABASE_URI
        cls._tmp_db = str(Path(__file__).resolve().parent / '.poller_concurrent_guard_test.db')
        if os.path.exists(cls._tmp_db):
            os.remove(cls._tmp_db)
        cfg_mod.TestingConfig.SQLALCHEMY_DATABASE_URI = f'sqlite:///{cls._tmp_db}'
        os.environ['FLASK_ENV'] = 'testing'
        from anetbbs.web_app import create_app
        from anetbbs.models import db
        cls.app = create_app('testing')
        cls.app.config['TESTING'] = True
        with cls.app.app_context():
            db.create_all()

    @classmethod
    def tearDownClass(cls):
        cfg_mod.TestingConfig.SQLALCHEMY_DATABASE_URI = cls._orig_db_uri
        for suffix in ('', '-wal', '-shm'):
            path = cls._tmp_db + suffix
            if os.path.exists(path):
                os.remove(path)

    def test_second_poll_is_skipped_while_first_is_still_running(self):
        from anetbbs.models import db, EchomailNetwork, EchomailPollLog
        from anetbbs.echomail import poller

        with self.app.app_context():
            net = EchomailNetwork(
                name='ConcurrentGuardNet', network_type='binkp',
                binkp_host='127.0.0.1', binkp_port=24554,
                our_address='1:1/20', hub_address='1:1/21',
                is_active=True, poll_interval_minutes=30)
            db.session.add(net)
            db.session.commit()

            # Simulate a poll already in progress (exactly the row
            # _do_poll() itself writes at the top of a real attempt).
            existing = EchomailPollLog(
                network_id=net.id, poll_type='both', status='running')
            db.session.add(existing)
            db.session.commit()

            called = {'n': 0}

            def _fake_run_client(*a, **kw):
                called['n'] += 1
                return {'received': [], 'sent': 0}

            with patch.object(poller, '_run_client', _fake_run_client):
                poller._do_poll(self.app, net)

            self.assertEqual(called['n'], 0,
                             'a second poll must not even attempt a connection '
                             'while one is already in progress for this network')
            # Still exactly one row -- no new EchomailPollLog created for
            # the skipped attempt.
            self.assertEqual(
                EchomailPollLog.query.filter_by(network_id=net.id).count(), 1)

    def test_poll_proceeds_normally_when_nothing_is_running(self):
        """Sanity check: the guard must not block the normal case."""
        from anetbbs.models import db, EchomailNetwork, EchomailPollLog
        from anetbbs.echomail import poller

        with self.app.app_context():
            net = EchomailNetwork(
                name='ConcurrentGuardNormalNet', network_type='binkp',
                binkp_host='127.0.0.1', binkp_port=24554,
                our_address='1:1/22', hub_address='1:1/23',
                is_active=True, poll_interval_minutes=30)
            db.session.add(net)
            db.session.commit()

            called = {'n': 0}

            def _fake_run_client(*a, **kw):
                called['n'] += 1
                return {'received': [], 'sent': 0}

            with patch.object(poller, '_run_client', _fake_run_client):
                poller._do_poll(self.app, net)

            self.assertEqual(called['n'], 1)
            self.assertEqual(
                EchomailPollLog.query.filter_by(network_id=net.id).count(), 1)
            row = EchomailPollLog.query.filter_by(network_id=net.id).first()
            self.assertEqual(row.status, 'success')

    def test_completed_prior_poll_does_not_block_a_new_one(self):
        """A row that's already moved past 'running' (success/error/
        partial) is old history, not a signal to skip."""
        from anetbbs.models import db, EchomailNetwork, EchomailPollLog
        from anetbbs.echomail import poller

        with self.app.app_context():
            net = EchomailNetwork(
                name='ConcurrentGuardCompletedNet', network_type='binkp',
                binkp_host='127.0.0.1', binkp_port=24554,
                our_address='1:1/24', hub_address='1:1/25',
                is_active=True, poll_interval_minutes=30)
            db.session.add(net)
            db.session.commit()
            db.session.add(EchomailPollLog(
                network_id=net.id, poll_type='both', status='success'))
            db.session.commit()

            called = {'n': 0}

            def _fake_run_client(*a, **kw):
                called['n'] += 1
                return {'received': [], 'sent': 0}

            with patch.object(poller, '_run_client', _fake_run_client):
                poller._do_poll(self.app, net)

            self.assertEqual(called['n'], 1)

    def test_stale_running_row_is_auto_recovered_not_skipped_forever(self):
        """Real bug found live: a poll interrupted mid-flight (e.g. a
        service restart) leaves its row stuck at status='running'
        forever -- confirmed live, DOVE-Net's poll history just went
        silent for over a day with a single old 'running' row and
        nothing after it, since the dedup guard treated that stale row
        as a genuine in-progress poll and skipped every attempt since.
        A 'running' row old enough to be implausible must be treated as
        abandoned, not as a permanent lock."""
        from anetbbs.models import db, EchomailNetwork, EchomailPollLog
        from anetbbs.echomail import poller

        with self.app.app_context():
            net = EchomailNetwork(
                name='ConcurrentGuardStaleNet', network_type='qwk',
                qwk_host='example.test', qwk_port=80,
                is_active=True, poll_interval_minutes=30)
            db.session.add(net)
            db.session.commit()

            stale_started = datetime.utcnow() - timedelta(
                minutes=poller._STALE_RUNNING_POLL_MINUTES + 5)
            stale_row = EchomailPollLog(
                network_id=net.id, poll_type='both', status='running',
                started_at=stale_started)
            db.session.add(stale_row)
            db.session.commit()
            stale_row_id = stale_row.id

            called = {'n': 0}

            def _fake_run_client(*a, **kw):
                called['n'] += 1
                return {'received': [], 'sent': 0}

            with patch.object(poller, '_run_client', _fake_run_client):
                poller._do_poll(self.app, net)

            self.assertEqual(
                called['n'], 1,
                'a poll blocked only by a long-stale "running" row must '
                'still proceed instead of being skipped forever')

            recovered = EchomailPollLog.query.get(stale_row_id)
            self.assertEqual(
                recovered.status, 'error',
                'the abandoned stale row must be flipped out of '
                '"running" so it stops masquerading as in-progress')

            new_row = (EchomailPollLog.query
                       .filter_by(network_id=net.id)
                       .filter(EchomailPollLog.id != stale_row_id)
                       .first())
            self.assertIsNotNone(new_row, 'a new poll row must have been created')
            self.assertEqual(new_row.status, 'success')

    def test_recently_running_row_still_blocks_normally(self):
        """Sanity check the staleness fix doesn't weaken the original
        guard -- a row that's genuinely only seconds old must still
        block a second concurrent attempt."""
        from anetbbs.models import db, EchomailNetwork, EchomailPollLog
        from anetbbs.echomail import poller

        with self.app.app_context():
            net = EchomailNetwork(
                name='ConcurrentGuardFreshNet', network_type='binkp',
                binkp_host='127.0.0.1', binkp_port=24554,
                our_address='1:1/26', hub_address='1:1/27',
                is_active=True, poll_interval_minutes=30)
            db.session.add(net)
            db.session.commit()
            db.session.add(EchomailPollLog(
                network_id=net.id, poll_type='both', status='running',
                started_at=datetime.utcnow() - timedelta(minutes=1)))
            db.session.commit()

            called = {'n': 0}

            def _fake_run_client(*a, **kw):
                called['n'] += 1
                return {'received': [], 'sent': 0}

            with patch.object(poller, '_run_client', _fake_run_client):
                poller._do_poll(self.app, net)

            self.assertEqual(called['n'], 0)
            self.assertEqual(
                EchomailPollLog.query.filter_by(network_id=net.id).count(), 1)


if __name__ == '__main__':
    unittest.main()



"""Regression tests for MRC bridge debug-log password redaction --
real gap found in a security/performance audit: MRC_BRIDGE_LOG_LEVEL=
DEBUG's raw wire-packet trace ("MRC RAW OUT/IN" in main.py, "MYSTIC RAW
OUT/IN" in mystic_connection.py) logged every packet completely
unredacted. IDENTIFY/REGISTER/UPDATE/ROOMPASS carry their password as
plain text inside the packet's own command field (per the real MRC
protocol spec template, e.g. "user~bbs~room~SERVER~msgext~~IDENTIFY
password~"), so enabling DEBUG tracing for any other diagnostic reason
put every login password a user typed into the bridge's own log file
in plaintext, permanently. Fixed via MRCProtocol.redact_packet_for_logs(),
now called at every raw-trace log site in both main.py and
mystic_connection.py.
"""
import sys
import unittest
from pathlib import Path

sys.path.insert(0, str(Path(__file__).resolve().parents[1]))

from mrc.bridge.mrc_protocol import MRCProtocol


class RedactCommandForLogsTests(unittest.TestCase):
    def test_identify_password_is_redacted(self):
        self.assertEqual(MRCProtocol.redact_command_for_logs('IDENTIFY hunter2'),
                         'IDENTIFY ********')

    def test_register_password_is_redacted(self):
        self.assertEqual(MRCProtocol.redact_command_for_logs('REGISTER hunter2'),
                         'REGISTER ********')

    def test_update_password_is_redacted(self):
        self.assertEqual(MRCProtocol.redact_command_for_logs('UPDATE newpass123'),
                         'UPDATE ********')

    def test_roompass_is_redacted(self):
        self.assertEqual(MRCProtocol.redact_command_for_logs('ROOMPASS secretroom'),
                         'ROOMPASS ********')

    def test_case_insensitive_match(self):
        # Matching is case-insensitive; the redacted marker itself is
        # always uppercase regardless of the input's original casing --
        # redaction doesn't need to preserve it, only hide the secret.
        self.assertEqual(MRCProtocol.redact_command_for_logs('identify hunter2'),
                         'IDENTIFY ********')

    def test_ordinary_command_is_untouched(self):
        self.assertEqual(MRCProtocol.redact_command_for_logs('MOTD'), 'MOTD')

    def test_ordinary_chat_message_is_untouched(self):
        self.assertEqual(MRCProtocol.redact_command_for_logs('hey everyone!'),
                         'hey everyone!')

    def test_empty_and_none_are_handled(self):
        self.assertEqual(MRCProtocol.redact_command_for_logs(''), '')
        self.assertEqual(MRCProtocol.redact_command_for_logs(None), '')


class RedactPacketForLogsTests(unittest.TestCase):
    def test_identify_password_removed_from_full_packet(self):
        packet = 'alice~bbs1~lobby~SERVER~~~IDENTIFY hunter2~'
        redacted = MRCProtocol.redact_packet_for_logs(packet)
        self.assertNotIn('hunter2', redacted)
        self.assertIn('IDENTIFY ********', redacted)
        # Everything else about the packet must still be visible for
        # real debugging -- only the password itself is masked.
        self.assertIn('alice', redacted)
        self.assertIn('bbs1', redacted)
        self.assertIn('lobby', redacted)

    def test_register_password_removed_from_full_packet(self):
        packet = 'bob~bbs1~lobby~SERVER~~~REGISTER newpass123~'
        redacted = MRCProtocol.redact_packet_for_logs(packet)
        self.assertNotIn('newpass123', redacted)
        self.assertIn('REGISTER ********', redacted)

    def test_ordinary_chat_packet_is_unaffected(self):
        packet = 'alice~bbs1~lobby~~~lobby~just chatting~'
        redacted = MRCProtocol.redact_packet_for_logs(packet)
        self.assertEqual(redacted, packet)

    def test_field_count_matches_original_after_redaction(self):
        packet = 'alice~bbs1~lobby~SERVER~~~IDENTIFY hunter2~'
        redacted = MRCProtocol.redact_packet_for_logs(packet)
        self.assertEqual(redacted.count('~'), packet.count('~'))

    def test_too_short_to_be_a_real_packet_is_returned_unchanged(self):
        garbage = 'not~a~real~packet'
        self.assertEqual(MRCProtocol.redact_packet_for_logs(garbage), garbage)

    def test_empty_and_none_are_handled(self):
        self.assertEqual(MRCProtocol.redact_packet_for_logs(''), '')
        self.assertIsNone(MRCProtocol.redact_packet_for_logs(None))


class RawTraceLogSitesUseRedactionTests(unittest.TestCase):
    """Guards against the exact regression this audit found: confirms
    every raw-packet DEBUG log call site actually routes through
    redact_packet_for_logs(), not just that the helper itself works in
    isolation -- a future refactor could easily reintroduce a bare
    logger.debug("...RAW...", packet) call without it."""

    def test_main_py_raw_trace_sites_call_the_redaction_helper(self):
        import inspect
        from mrc.bridge import main as bridge_main
        source = inspect.getsource(bridge_main)
        for marker in ('logger.debug("MRC RAW OUT', 'logger.debug("MRC RAW IN'):
            idx = source.index(marker)
            line_end = source.index('\n', idx)
            line = source[idx:line_end]
            self.assertIn('redact_packet_for_logs', line,
                          f'{marker} log line must redact via redact_packet_for_logs()')

    def test_mystic_connection_py_raw_trace_sites_call_the_redaction_helper(self):
        import inspect
        from mrc.bridge import mystic_connection
        source = inspect.getsource(mystic_connection)
        for marker in ('logger.debug("MYSTIC RAW OUT', 'logger.debug("MYSTIC RAW IN'):
            idx = source.index(marker)
            line_end = source.index('\n', idx)
            line = source[idx:line_end]
            self.assertIn('redact_packet_for_logs', line,
                          f'{marker} log line must redact via redact_packet_for_logs()')


if __name__ == '__main__':
    unittest.main()



"""Games section (anetbbs-cfg) -- door games, categories, active sessions.

Mirrors anetbbs/web/games_admin.py's GameForm field set. msgbase_area_id
(InterBBS score-sharing EchoArea) is a raw nullable FK id here rather
than a grouped picker like the web admin's dropdown -- check the
Echomail section for area ids. Same "known gap" pattern as other
sections' web-admin-only fields.
"""
from datetime import datetime

from sqlalchemy.exc import IntegrityError

from anetbbs.cfg import ui
from anetbbs.models import db, Game, GameCategory, GameSession

GAME_TYPE_CHOICES = [
    "builtin_web", "door_dos_browser", "door_dos", "door_dosemu",
    "door_mystic", "door_mystic_mps", "door_native", "door_synchronet",
    "door_rlogin", "door_telnet",
]

DROP_FILE_CHOICES = ["none", "door.sys", "dorinfo", "door32.sys"]

GAME_FIELDS = [
    {"key": "name", "label": "Name", "kind": "text"},
    {"key": "slug", "label": "Slug (URL-safe)", "kind": "text"},
    {"key": "description", "label": "Description", "kind": "text_nullable"},
    {"key": "category", "label": "Category (slug)", "kind": "text_nullable"},
    {"key": "min_access_level", "label": "Min Access Level", "kind": "int"},
    {"key": "game_type", "label": "Game Type", "kind": "choice", "choices": GAME_TYPE_CHOICES},
    {"key": "icon", "label": "Bootstrap Icon Class", "kind": "text_nullable"},
    {"key": "max_nodes", "label": "Max Simultaneous Players", "kind": "int"},
    {"key": "sort_order", "label": "Sort Order", "kind": "int"},
    {"key": "is_active", "label": "Active", "kind": "bool"},
    {"key": "is_multiplayer", "label": "Multiplayer", "kind": "bool"},
    {"key": "web_enabled", "label": "Web Enabled", "kind": "bool"},
    {"key": "terminal_enabled", "label": "Terminal Enabled", "kind": "bool"},
    {"key": "share_scores_interbbs", "label": "Share Scores InterBBS", "kind": "bool"},
    {"key": "executable_path", "label": "Executable Path", "kind": "text_nullable"},
    {"key": "working_directory", "label": "Working Directory", "kind": "text_nullable"},
    {"key": "command_line_args", "label": "Command Line Args", "kind": "text_nullable"},
    {"key": "rlogin_bbs_tag", "label": "rlogin BBS Tag", "kind": "text_nullable"},
    {"key": "drop_file_type", "label": "Drop File Type", "kind": "choice", "choices": DROP_FILE_CHOICES},
    {"key": "drop_file_path", "label": "Drop File Path", "kind": "text_nullable"},
    {"key": "use_dosbox", "label": "Use DOSBox/dosemu2", "kind": "bool"},
    {"key": "needs_fossil_driver", "label": "Needs FOSSIL Driver", "kind": "bool"},
    {"key": "mystic_script_path", "label": "Mystic Script Path (.mpy)", "kind": "text_nullable"},
    {"key": "synchronet_script_path", "label": "Synchronet Script Path (.js)", "kind": "text_nullable"},
    {"key": "synchronet_exec_dir", "label": "Synchronet Exec Dir", "kind": "text_nullable"},
    {"key": "msgbase_area_id", "label": "InterBBS Score-Sharing Area ID", "kind": "int_nullable"},
    {"key": "web_game_module", "label": "Web Game Module Name", "kind": "text_nullable"},
    {"key": "web_game_url", "label": "Web Game URL", "kind": "text_nullable"},
]

GAME_HELP = [
    "InterBBS Score-Sharing Area ID: raw EchoArea id (see Echomail section); blank = off.",
]

GAME_NEW_DEFAULTS = {
    "name": "", "slug": "", "description": None, "category": "other",
    "min_access_level": 0, "game_type": "door_dos", "icon": None,
    "max_nodes": 1, "sort_order": 0, "is_active": True, "is_multiplayer": False,
    "web_enabled": True, "terminal_enabled": True, "share_scores_interbbs": True,
    "executable_path": None, "working_directory": None, "command_line_args": None,
    "rlogin_bbs_tag": None, "drop_file_type": "none", "drop_file_path": None,
    "use_dosbox": False, "needs_fossil_driver": False, "mystic_script_path": None,
    "synchronet_script_path": None, "synchronet_exec_dir": None,
    "msgbase_area_id": None, "web_game_module": None, "web_game_url": None,
}

GAME_COLUMNS = [
    ("Order", 6, lambda g: g.sort_order),
    ("Name", 22, lambda g: g.name),
    ("Type", 16, lambda g: g.game_type),
    ("Category", 12, lambda g: g.category or ""),
    ("Active", 6, lambda g: "Yes" if g.is_active else "No"),
]

CATEGORY_FIELDS = [
    {"key": "name", "label": "Name", "kind": "text"},
    {"key": "slug", "label": "Slug (URL-safe)", "kind": "text"},
    {"key": "sort_order", "label": "Sort Order", "kind": "int"},
    {"key": "as_submenu", "label": "Show As Submenu Section", "kind": "bool"},
]

CATEGORY_NEW_DEFAULTS = {"name": "", "slug": "", "sort_order": 0, "as_submenu": False}

CATEGORY_COLUMNS = [
    ("Order", 6, lambda c: c.sort_order),
    ("Name", 24, lambda c: c.name),
    ("Slug", 24, lambda c: c.slug),
    ("Submenu", 8, lambda c: "Yes" if c.as_submenu else "No"),
]

SESSION_COLUMNS = [
    ("ID", 6, lambda s: s.id),
    ("Game", 20, lambda s: s.game.name if s.game else "?"),
    ("User", 16, lambda s: s.user.username if s.user else "?"),
    ("Node", 5, lambda s: s.node_number),
    ("Started", 20, lambda s: s.started_at.strftime("%Y-%m-%d %H:%M") if s.started_at else ""),
]


def list_games():
    return Game.query.order_by(Game.sort_order, Game.name).all()


def values_from_game(g):
    return {f["key"]: getattr(g, f["key"]) for f in GAME_FIELDS}


def create_game(data):
    g = Game(**data)
    db.session.add(g)
    db.session.commit()
    return g


def update_game(g, data):
    for k, v in data.items():
        setattr(g, k, v)
    db.session.commit()


def delete_game(g):
    db.session.delete(g)
    db.session.commit()


def list_categories():
    return GameCategory.query.order_by(GameCategory.sort_order, GameCategory.name).all()


def values_from_category(c):
    return {f["key"]: getattr(c, f["key"]) for f in CATEGORY_FIELDS}


def create_category(data):
    c = GameCategory(**data)
    db.session.add(c)
    db.session.commit()
    return c


def update_category(c, data):
    for k, v in data.items():
        setattr(c, k, v)
    db.session.commit()


def delete_category(c):
    db.session.delete(c)
    db.session.commit()


def list_active_sessions():
    return (GameSession.query.filter_by(status="active")
            .order_by(GameSession.started_at.desc()).all())


def disconnect_session(session_id):
    from anetbbs.games.door_runner import terminate_session
    terminate_session(session_id)
    gs = GameSession.query.get(session_id)
    if gs:
        gs.status = "completed"
        gs.ended_at = datetime.utcnow()
        db.session.commit()


def clear_stale_sessions():
    count = GameSession.query.filter_by(status="active").update(
        {"status": "stale", "ended_at": datetime.utcnow()})
    db.session.commit()
    return count


def _add_game(stdscr):
    data = ui.run_form(stdscr, "New Game", GAME_FIELDS, dict(GAME_NEW_DEFAULTS), help_lines=GAME_HELP)
    if data is None:
        return
    if not data.get("name") or not data.get("slug"):
        ui.show_message(stdscr, "Name and Slug are required.", error=True)
        return
    try:
        create_game(data)
    except IntegrityError:
        db.session.rollback()
        ui.show_message(stdscr, "A game with that slug already exists.", error=True)


def _edit_game(stdscr, g):
    data = ui.run_form(stdscr, f"Edit Game: {g.name}", GAME_FIELDS, values_from_game(g), help_lines=GAME_HELP)
    if data is None:
        return
    try:
        update_game(g, data)
    except IntegrityError:
        db.session.rollback()
        ui.show_message(stdscr, "A game with that slug already exists.", error=True)


def _delete_game(stdscr, g):
    if ui.confirm(stdscr, f"Delete game '{g.name}'?"):
        delete_game(g)


def _run_games(stdscr):
    ui.run_list(
        stdscr, "Games", GAME_COLUMNS, list_games,
        on_add=_add_game, on_edit=_edit_game, on_delete=_delete_game,
    )


def _add_category(stdscr):
    data = ui.run_form(stdscr, "New Category", CATEGORY_FIELDS, dict(CATEGORY_NEW_DEFAULTS))
    if data is None:
        return
    if not data.get("name") or not data.get("slug"):
        ui.show_message(stdscr, "Name and Slug are required.", error=True)
        return
    try:
        create_category(data)
    except IntegrityError:
        db.session.rollback()
        ui.show_message(stdscr, "A category with that slug already exists.", error=True)


def _edit_category(stdscr, c):
    data = ui.run_form(stdscr, f"Edit Category: {c.name}", CATEGORY_FIELDS, values_from_category(c))
    if data is None:
        return
    try:
        update_category(c, data)
    except IntegrityError:
        db.session.rollback()
        ui.show_message(stdscr, "A category with that slug already exists.", error=True)


def _delete_category(stdscr, c):
    if ui.confirm(stdscr, f"Delete category '{c.name}'?"):
        delete_category(c)


def _run_categories(stdscr):
    ui.run_list(
        stdscr, "Game Categories", CATEGORY_COLUMNS, list_categories,
        on_add=_add_category, on_edit=_edit_category, on_delete=_delete_category,
    )


def _disconnect(stdscr, s):
    if ui.confirm(stdscr, f"Disconnect session #{s.id} ({s.user.username if s.user else '?'})?"):
        disconnect_session(s.id)


def _clear_stale(stdscr, _row):
    if ui.confirm(stdscr, "Mark ALL active sessions as stale?"):
        count = clear_stale_sessions()
        ui.show_message(stdscr, f"Marked {count} session(s) as stale.")


def _run_sessions(stdscr):
    ui.run_list(
        stdscr, "Active Game Sessions", SESSION_COLUMNS, list_active_sessions,
        on_delete=_disconnect,
        extra_actions={"c": ("ClearStale", _clear_stale)},
        empty_hint="(no active sessions)",
    )


def run(stdscr):
    items = [("games", "Games"), ("categories", "Categories"), ("sessions", "Active Sessions")]
    while True:
        choice = ui.run_menu(stdscr, "Games", items)
        if choice is None:
            return
        if choice == "games":
            _run_games(stdscr)
        elif choice == "categories":
            _run_categories(stdscr)
        elif choice == "sessions":
            _run_sessions(stdscr)



"""Regression test for a real privilege-escalation bug found in a
security audit: anetbbs-cfg (the standalone curses tool that can edit
user security levels, echomail/hub credentials, etc.) is only ever
supposed to be reachable SSH-only + admin-only. bbs_ui.py's own
_sysop_cfg_tool() docstring specifically claimed it was "gated twice"
so a bug elsewhere in the menu system couldn't bypass that -- but
menu_engine.py's _act_door() (the generic "launch a door by raw
numeric id" menu action, reachable via any sysop-configured menu item)
performed NO admin/SSH/is_active check at all, and reached
door_runner.py's play_door_game_telnet() -- the real, shared door-
launch entry point -- directly, bypassing bbs_ui.py's gating entirely.

Two fixes, tested here:
  1. menu_engine._act_door() now refuses any is_active=False game
     (general hygiene -- hidden games should never be reachable by a
     generic numeric-id menu action).
  2. door_runner.play_door_game_telnet() independently refuses to
     launch the specific 'anetbbs-cfg' game unless the session is SSH
     AND the user is admin -- enforced at the actual shared launch
     chokepoint, not just in bbs_ui.py's wrapper, so no current or
     future caller can bypass it.
"""
import asyncio
import os
import sys
import types
import unittest
from pathlib import Path
from unittest.mock import AsyncMock, patch

sys.path.insert(0, str(Path(__file__).resolve().parents[1]))

import anetbbs.config as cfg_mod


class _FakeSession:
    def __init__(self, protocol=None, user=None):
        self.written = []
        self.encoding = 'cp437'
        self.presence = types.SimpleNamespace(protocol=protocol) if protocol else None
        self.user = user or {}

    async def write(self, data):
        self.written.append(data)

    async def read_line(self, prompt=''):
        if prompt:
            await self.write(prompt)
        return ''

    def transcript(self):
        return ''.join(self.written)


class AnetbbsCfgPrivilegeEscalationTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._orig_db_uri = cfg_mod.TestingConfig.SQLALCHEMY_DATABASE_URI
        cls._tmp_db = str(Path(__file__).resolve().parent / '.cfg_privesc_test.db')
        if os.path.exists(cls._tmp_db):
            os.remove(cls._tmp_db)
        cfg_mod.TestingConfig.SQLALCHEMY_DATABASE_URI = f'sqlite:///{cls._tmp_db}'
        os.environ['FLASK_ENV'] = 'testing'

        from anetbbs.web_app import create_app
        from anetbbs.models import db, User, Game
        cls.app = create_app('testing')
        cls.app.config['TESTING'] = True
        with cls.app.app_context():
            db.create_all()

            regular = User(username='regularuser', email='reg@example.com',
                            password_hash='x', access_level=50, is_admin=False)
            admin = User(username='sysop', email='sysop@example.com',
                         password_hash='x', access_level=100, is_admin=True)
            db.session.add_all([regular, admin])

            # create_app() itself bundled-door-seeds a REAL Game row at
            # slug='anetbbs-cfg' (web_app.py), but only when the
            # anetbbs-cfg console-script binary actually exists on
            # disk (its own 'must_exist' gate) -- true in a real
            # install/Docker build (setup.py's console_scripts entry
            # point gets installed for real), false in a bare
            # `pip install -e .` dev checkout with no such binary
            # present. A blind INSERT here collided with that real
            # seed row's UNIQUE(slug) constraint the moment this ran
            # somewhere the binary DOES exist (caught by CI, not by
            # local dev testing) -- get-or-create instead, so this
            # test's own outcome doesn't depend on which of those two
            # environments it happens to run in.
            cfg_game = Game.query.filter_by(slug='anetbbs-cfg').first()
            if cfg_game is None:
                cfg_game = Game(name='Config Tool', slug='anetbbs-cfg',
                                game_type='door_native', is_active=False,
                                web_enabled=False,
                                executable_path='/bin/true')
                db.session.add(cfg_game)
            else:
                cfg_game.is_active = False
            hidden_other = Game(name='Some Other Hidden Door', slug='hidden-other',
                                game_type='door_native', is_active=False,
                                executable_path='/bin/true')
            visible = Game(name='Visible Door', slug='visible-door',
                           game_type='door_native', is_active=True,
                           executable_path='/bin/true')
            db.session.add_all([hidden_other, visible])
            db.session.commit()

            cls.regular_id = regular.id
            cls.admin_id = admin.id
            cls.cfg_game_id = cfg_game.id
            cls.hidden_other_id = hidden_other.id
            cls.visible_id = visible.id

    @classmethod
    def tearDownClass(cls):
        cfg_mod.TestingConfig.SQLALCHEMY_DATABASE_URI = cls._orig_db_uri
        if os.path.exists(cls._tmp_db):
            os.remove(cls._tmp_db)

    def _app_ctx(self):
        return patch('anetbbs.features.bbs_ui._app', return_value=self.app)

    # ── Fix 1: menu_engine._act_door() refuses hidden (is_active=False) games ──

    def test_act_door_refuses_the_cfg_tool_by_id(self):
        from anetbbs.features.menu_engine import _act_door

        class _UI:
            def __init__(self, session):
                self.session = session

        session = _FakeSession(protocol='ssh', user={'id': self.admin_id, 'is_admin': True})
        ui = _UI(session)
        with self._app_ctx(), patch('anetbbs.games.door_runner.play_door_game_telnet',
                                     new=AsyncMock()) as mock_launch:
            asyncio.run(_act_door(ui, str(self.cfg_game_id)))
        mock_launch.assert_not_awaited()
        self.assertIn('Game not found', session.transcript())

    def test_act_door_refuses_any_hidden_game_not_just_cfg(self):
        from anetbbs.features.menu_engine import _act_door

        class _UI:
            def __init__(self, session):
                self.session = session

        session = _FakeSession(protocol='telnet', user={'id': self.regular_id})
        ui = _UI(session)
        with self._app_ctx(), patch('anetbbs.games.door_runner.play_door_game_telnet',
                                     new=AsyncMock()) as mock_launch:
            asyncio.run(_act_door(ui, str(self.hidden_other_id)))
        mock_launch.assert_not_awaited()

    def test_act_door_still_launches_a_normal_visible_game(self):
        from anetbbs.features.menu_engine import _act_door

        class _UI:
            def __init__(self, session):
                self.session = session

        session = _FakeSession(protocol='telnet', user={'id': self.regular_id})
        ui = _UI(session)
        with self._app_ctx(), patch('anetbbs.games.door_runner.play_door_game_telnet',
                                     new=AsyncMock()) as mock_launch:
            asyncio.run(_act_door(ui, str(self.visible_id)))
        mock_launch.assert_awaited_once()

    # ── Fix 2: play_door_game_telnet() itself refuses anetbbs-cfg for
    #    non-SSH or non-admin sessions, independent of the caller ──

    def test_play_door_game_telnet_refuses_cfg_over_telnet_even_for_admin(self):
        from anetbbs.games.door_runner import play_door_game_telnet
        from anetbbs.models import Game
        with self.app.app_context():
            game = Game.query.get(self.cfg_game_id)
            result = asyncio.run(play_door_game_telnet(
                game, {'id': self.admin_id}, _FakeSession(protocol='telnet')))
        self.assertFalse(result)

    def test_play_door_game_telnet_refuses_cfg_over_ssh_for_non_admin(self):
        from anetbbs.games.door_runner import play_door_game_telnet
        from anetbbs.models import Game
        with self.app.app_context():
            game = Game.query.get(self.cfg_game_id)
            result = asyncio.run(play_door_game_telnet(
                game, {'id': self.regular_id}, _FakeSession(protocol='ssh')))
        self.assertFalse(result)

    def test_play_door_game_telnet_refuses_cfg_with_no_session_presence_at_all(self):
        # Simulates a caller that never went through the normal
        # login/protocol-detection flow at all -- the most direct
        # possible bypass attempt.
        from anetbbs.games.door_runner import play_door_game_telnet
        from anetbbs.models import Game
        with self.app.app_context():
            game = Game.query.get(self.cfg_game_id)
            result = asyncio.run(play_door_game_telnet(
                game, {'id': self.admin_id}, _FakeSession(protocol=None)))
        self.assertFalse(result)


if __name__ == '__main__':
    unittest.main()



# anetbbs/features/lastcallers.py
"""Last Callers -- read-only paginated list of recent BBS logins.

Deliberately mirrors menu_engine.py's existing `_act_oneliners` inline
"Last 10 Callers" block (plain fixed-width list, not Wall's box-drawing
chrome) since the sysop already has a visual precedent on this BBS for
what a callers list looks like -- this screen is meant to coexist with
that inline block as a fuller, paginated superset (filterable by
service, more rows), not a differently-styled reskin of the same data.

Can optionally show entries imported via InterBBS Last Callers sharing
(anetbbs/echomail/interbbs_sync.py) tagged with their origin BBS name.
"""
from .bbs_ui import _app
from ..core.tz import fmt_eastern

PAGE_SIZE = 20


def recent_callers_query(limit=200):
    """Shared query builder for every user-facing Last Callers display
    (terminal full view, terminal inline block, web oneliners page --
    NOT the admin audit view, which always shows everything).

    Applies the LASTCALLERS_HIDE_SYSOP filter: excludes CallerLog rows
    for a local login by an is_admin user. Uses an outer join so rows
    with no local user_id at all (InterBBS-imported entries, which
    never carry a real user_id -- see interbbs_sync.py's
    sync_lastcallers_inbound) are never affected by this filter; there's
    no local User to check for those, and a sysop hiding their own
    frequent local logins says nothing about a remote BBS's callers.
    """
    from flask import current_app
    from ..models import db, CallerLog, User

    query = CallerLog.query.order_by(CallerLog.started_at.desc())
    if current_app.config.get('LASTCALLERS_HIDE_SYSOP'):
        # isnot(True), not is_(False) -- a legacy/migrated User row with
        # is_admin IS NULL (nullable column, no DB-level default) must
        # read as "not admin", not get excluded by SQL's three-valued
        # NULL logic.
        query = (query.outerjoin(User, CallerLog.user_id == User.id)
                 .filter(db.or_(User.is_admin.isnot(True), User.id.is_(None))))
    return query.limit(limit)


async def show_last_callers(session, args=None):
    """Entry point -- takes `session` directly (not the full BBSMenuUI),
    matching wall.py's show_wall(session, ...) signature, since this is
    called both from menu_engine.py's _act_lastcallers (which has a `ui`
    to pull .session off of) and login_modules.py's _dispatch (which
    only ever has `session`, no `ui`)."""
    with _app().app_context():
        from flask import current_app
        display_count = current_app.config.get('LASTCALLERS_DISPLAY_COUNT', 20)
        callers = recent_callers_query(display_count).all()
        rows = [
            (c.username or '?', c.service or '?', c.started_at, c.origin_bbs)
            for c in callers
        ]

    await session.write('\x1b[2J\x1b[H')

    if not rows:
        await session.write("\x1b[1;36m=== Last Callers ===\x1b[0m\r\n"
                            "  (none yet)\r\n")
        await session.read_line("\r\nPress Enter...")
        return

    lines = ["\x1b[1;36m=== Last Callers ===\x1b[0m"]
    for username, service, started_at, origin_bbs in rows:
        when = fmt_eastern(started_at, '%m-%d %H:%M', '?')
        origin_tag = f" \x1b[36m(via {origin_bbs})\x1b[0m" if origin_bbs else ""
        lines.append(
            f"  \x1b[1;33m{username:<20}\x1b[0m \x1b[1;30m{service:<8}\x1b[0m "
            f"{when}{origin_tag}")

    await _page_lines(session, lines, page_size=PAGE_SIZE,
                      end_message='--- end of last callers ---')


async def _page_lines(session, lines, page_size=22, end_message='--- end ---'):
    """Standalone pager, same shape as BBSMenuUI._page_lines() in
    bbs_ui.py, but usable without a `ui` instance (only `session`)."""
    YEL = '\x1b[1;33m'; DIM = '\x1b[37m'; RESET = '\x1b[0m'
    total = len(lines)
    if not total:
        return
    page = 0
    while page * page_size < total:
        chunk = lines[page * page_size:(page + 1) * page_size]
        for line in chunk:
            await session.write(line + '\r\n')
        page += 1
        if page * page_size >= total:
            await session.read_line(
                f'{YEL}{end_message}{RESET}  Press Enter to continue: ')
            return
        shown = min(page * page_size, total)
        prompt = (f'{YEL}--MORE--{RESET}  '
                 f'({shown}/{total} lines)  '
                 f'{DIM}[Enter]=more  [Q]=quit:{RESET} ')
        ans = (await session.read_line(prompt) or '').strip().upper()
        if ans == 'Q':
            return



"""Route tests for HubIdentity admin CRUD (anetbbs/web/hub_admin.py's
/admin/echomail/hub/identities/* routes) -- Phase 1 of the
multi-hub-identity feature. Purely additive: nothing else in the app
consumes HubIdentity yet, so these tests only cover the CRUD itself
(create, edit, set-default invariant, delete guards).
"""
import os
import shutil
import sys
import tempfile
import unittest
from pathlib import Path

sys.path.insert(0, str(Path(__file__).resolve().parents[1]))

_DATA_DIR = Path(__file__).resolve().parents[1] / 'data'


def _snapshot_data_dir():
    if not _DATA_DIR.is_dir():
        return set()
    return set(_DATA_DIR.iterdir())


def _make_app(db_path):
    import anetbbs.config as cfg_mod
    if os.path.exists(db_path):
        os.remove(db_path)
    cfg_mod.TestingConfig.SQLALCHEMY_DATABASE_URI = f'sqlite:///{db_path}'
    os.environ['FLASK_ENV'] = 'testing'

    from anetbbs.web_app import create_app
    app = create_app('testing')
    app.config['TESTING'] = True
    app.config['WTF_CSRF_ENABLED'] = False
    app.config['REGISTRY_MODE_ENABLED'] = True
    return app


class HubIdentityCrudTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._data_dir_before = _snapshot_data_dir()
        import anetbbs.config as cfg_mod
        cls._orig_db_uri = cfg_mod.TestingConfig.SQLALCHEMY_DATABASE_URI
        cls._orig_flask_env = os.environ.get('FLASK_ENV')

    @classmethod
    def tearDownClass(cls):
        import anetbbs.config as cfg_mod
        cfg_mod.TestingConfig.SQLALCHEMY_DATABASE_URI = cls._orig_db_uri
        if cls._orig_flask_env is None:
            os.environ.pop('FLASK_ENV', None)
        else:
            os.environ['FLASK_ENV'] = cls._orig_flask_env
        for entry in _snapshot_data_dir() - cls._data_dir_before:
            if entry.is_dir():
                shutil.rmtree(entry, ignore_errors=True)
            else:
                entry.unlink(missing_ok=True)

    def setUp(self):
        self._tmp = tempfile.TemporaryDirectory()
        self.addCleanup(self._tmp.cleanup)

    def _login_as_admin(self, app, client):
        from anetbbs.models import db, User
        with app.app_context():
            admin = User.query.filter_by(username='admin').first()
            if admin is None:
                admin = User(username='admin', email='admin@example.com', is_admin=True)
                admin.set_password('password123')
                db.session.add(admin)
                db.session.commit()
            admin_id = admin.id
        with client.session_transaction() as sess:
            sess['_user_id'] = str(admin_id)
            sess['_fresh'] = True

    def test_create_second_identity(self):
        app = _make_app(str(Path(self._tmp.name) / 'a.db'))
        client = app.test_client()
        self._login_as_admin(app, client)

        resp = client.post('/admin/echomail/hub/identities/new', data={
            'name': 'Second Network', 'slug': 'secondnet',
            'qwk_hub_id': 'SECOND', 'binkp_zone': '2', 'binkp_net': '1',
            'binkp_hub_node': '1', 'binkp_domain': 'secnet',
            'nodelist_sysop': '', 'nodelist_location': '',
            'nodelist_phone': '', 'nodelist_speed': '115200',
            'is_active': 'y',
        }, follow_redirects=True)
        self.assertEqual(resp.status_code, 200)

        from anetbbs.models import HubIdentity
        with app.app_context():
            identities = HubIdentity.query.order_by(HubIdentity.name).all()
            self.assertEqual(len(identities), 2)
            second = HubIdentity.query.filter_by(slug='secondnet').first()
            self.assertIsNotNone(second)
            self.assertEqual(second.qwk_hub_id, 'SECOND')
            self.assertEqual(second.binkp_zone, 2)
            self.assertFalse(second.is_default,
                              'a newly created identity must not silently steal default')

    def test_duplicate_slug_rejected(self):
        app = _make_app(str(Path(self._tmp.name) / 'b.db'))
        client = app.test_client()
        self._login_as_admin(app, client)
        from anetbbs.models import HubIdentity
        with app.app_context():
            default_slug = HubIdentity.query.filter_by(is_default=True).one().slug

        resp = client.post('/admin/echomail/hub/identities/new', data={
            'name': 'Collides', 'slug': default_slug,
            'binkp_hub_node': '1', 'nodelist_speed': '115200',
        }, follow_redirects=True)
        self.assertEqual(resp.status_code, 200)
        with app.app_context():
            self.assertEqual(HubIdentity.query.count(), 1,
                              'duplicate slug must not create a second row')

    def test_set_default_moves_flag_off_old_default(self):
        app = _make_app(str(Path(self._tmp.name) / 'c.db'))
        client = app.test_client()
        self._login_as_admin(app, client)
        from anetbbs.models import db, HubIdentity
        with app.app_context():
            original_default_id = HubIdentity.query.filter_by(is_default=True).one().id
            second = HubIdentity(name='Second', slug='second2',
                                 binkp_zone=2, binkp_net=1, binkp_hub_node=1)
            db.session.add(second)
            db.session.commit()
            second_id = second.id

        resp = client.post(f'/admin/echomail/hub/identities/{second_id}/set-default',
                           follow_redirects=True)
        self.assertEqual(resp.status_code, 200)

        with app.app_context():
            defaults = HubIdentity.query.filter_by(is_default=True).all()
            self.assertEqual(len(defaults), 1, 'exactly one row must be default')
            self.assertEqual(defaults[0].id, second_id)
            old = HubIdentity.query.get(original_default_id)
            self.assertFalse(old.is_default)

    def test_cannot_delete_the_only_identity(self):
        app = _make_app(str(Path(self._tmp.name) / 'd.db'))
        client = app.test_client()
        self._login_as_admin(app, client)
        from anetbbs.models import HubIdentity
        with app.app_context():
            only_id = HubIdentity.query.one().id

        resp = client.post(f'/admin/echomail/hub/identities/{only_id}/delete',
                           follow_redirects=True)
        self.assertEqual(resp.status_code, 200)
        with app.app_context():
            self.assertEqual(HubIdentity.query.count(), 1)

    def test_cannot_delete_the_default_identity_even_with_others_present(self):
        app = _make_app(str(Path(self._tmp.name) / 'e.db'))
        client = app.test_client()
        self._login_as_admin(app, client)
        from anetbbs.models import db, HubIdentity
        with app.app_context():
            default_id = HubIdentity.query.filter_by(is_default=True).one().id
            second = HubIdentity(name='Second', slug='second3',
                                 binkp_zone=2, binkp_net=1, binkp_hub_node=1)
            db.session.add(second)
            db.session.commit()

        resp = client.post(f'/admin/echomail/hub/identities/{default_id}/delete',
                           follow_redirects=True)
        self.assertEqual(resp.status_code, 200)
        with app.app_context():
            self.assertIsNotNone(HubIdentity.query.get(default_id),
                                  'default identity must survive the delete attempt')
            self.assertEqual(HubIdentity.query.count(), 2)

    def test_cannot_delete_identity_with_nodes_assigned(self):
        app = _make_app(str(Path(self._tmp.name) / 'f.db'))
        client = app.test_client()
        self._login_as_admin(app, client)
        from anetbbs.models import db, HubIdentity, BinkPNode
        with app.app_context():
            second = HubIdentity(name='Second', slug='second4',
                                 binkp_zone=2, binkp_net=1, binkp_hub_node=1)
            db.session.add(second)
            db.session.commit()
            node = BinkPNode(name='Node A', ftn_address='2:1/100', password='x',
                             hub_identity_id=second.id)
            db.session.add(node)
            db.session.commit()
            second_id = second.id

        resp = client.post(f'/admin/echomail/hub/identities/{second_id}/delete',
                           follow_redirects=True)
        self.assertEqual(resp.status_code, 200)
        with app.app_context():
            self.assertIsNotNone(HubIdentity.query.get(second_id),
                                  'identity with a node still assigned must not be deletable')

    def test_delete_succeeds_once_unused_and_not_default(self):
        app = _make_app(str(Path(self._tmp.name) / 'g.db'))
        client = app.test_client()
        self._login_as_admin(app, client)
        from anetbbs.models import db, HubIdentity
        with app.app_context():
            second = HubIdentity(name='Unused', slug='unused',
                                 binkp_zone=3, binkp_net=1, binkp_hub_node=1)
            db.session.add(second)
            db.session.commit()
            second_id = second.id

        resp = client.post(f'/admin/echomail/hub/identities/{second_id}/delete',
                           follow_redirects=True)
        self.assertEqual(resp.status_code, 200)
        with app.app_context():
            self.assertIsNone(HubIdentity.query.get(second_id))


if __name__ == '__main__':
    unittest.main()
